@@ -1,9 +1,5 @@
 """Mach-O/ARM64 analyzer"""
 
-from .debug_util import DebugUtil
-
-__version__ = "10.5.4"
+__version__ = "13.0.5"
 __author__ = "Data Theorem"
 __url__ = "https://github.com/datatheorem/strongarm"
-
-__all__ = ["DebugUtil"]
