@@ -1,5 +1,4 @@
 import functools
-import logging
 import pathlib
 import shutil
 import sqlite3
@@ -8,13 +7,14 @@ import time
 from contextlib import closing
 from ctypes import sizeof
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, TypeVar, cast
+from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Type, TypeVar, cast
 
 from capstone import CS_ARCH_ARM64, CS_MODE_ARM, Cs, CsInsn
-from more_itertools import pairwise
+from more_itertools import first, pairwise
 
+from strongarm.logger import strongarm_logger
 from strongarm.macho.arch_independent_structs import CFString32, CFString64, CFStringStruct
-from strongarm.macho.dyld_info_parser import DyldBoundSymbol, DyldInfoParser
+from strongarm.macho.dyld_info_parser import DyldBoundSymbol
 from strongarm.macho.macho_binary import InvalidAddressError, MachoBinary
 from strongarm.macho.macho_definitions import VirtualMemoryPointer
 from strongarm.macho.macho_imp_stubs import MachoImpStubsParser
@@ -30,6 +30,7 @@ from strongarm.macho.objc_runtime_data_parser import (
 if TYPE_CHECKING:
     from strongarm.objc import ObjcFunctionAnalyzer, ObjcMethodInfo
 
+logger = strongarm_logger.getChild(__file__)
 
 _T = TypeVar("_T")
 
@@ -78,8 +79,7 @@ ANALYZER_SQL_SCHEMA = """
 
 
 class DisassemblyFailedError(Exception):
-    """Raised when Capstone fails to disassemble a bytecode sequence.
-    """
+    """Raised when Capstone fails to disassemble a bytecode sequence."""
 
 
 @dataclass(order=True, frozen=True)
@@ -97,8 +97,7 @@ class ObjcMsgSendXref(CallerXRef):
 
 @dataclass
 class CallableSymbol:
-    """A locally-defined function or externally-defined imported function
-    """
+    """A locally-defined function or externally-defined imported function."""
 
     address: VirtualMemoryPointer
     is_imported: bool
@@ -112,7 +111,9 @@ def _requires_xrefs_computed(func: CallableT) -> CallableT:
     @functools.wraps(func)
     def wrap(self: "MachoAnalyzer", *args: Any, **kwargs: Any) -> Any:
         if not self._has_computed_xrefs:
-            logging.info(f"called {func.__name__} before XRefs were computed, computing now...")
+            logger.info(
+                f"called {func.__name__} before XRefs were computed for {self.binary.path.name}, computing now..."
+            )
             self._build_xref_database()
         return func(self, *args, **kwargs)
 
@@ -128,13 +129,13 @@ class cached_property(object):
     https://github.com/pallets/werkzeug/blob/0e1b8c4fe598725b343085c5a9a867e90b966db6/werkzeug/utils.py#L35-L73
     """
 
-    def __init__(self, func, name=None, doc=None):
-        self.__name__ = name or func.__name__
+    def __init__(self, func: Callable) -> None:
+        self.__name__ = func.__name__
         self.__module__ = func.__module__
-        self.__doc__ = doc or func.__doc__
+        self.__doc__ = func.__doc__
         self.func = func
 
-    def __get__(self, obj, type=None):
+    def __get__(self, obj: Any, _type: Type = None) -> Any:
         if obj is None:
             return self
         value = obj.__dict__.get(self.__name__, None)
@@ -157,8 +158,6 @@ class MachoAnalyzer:
         self.cs = Cs(CS_ARCH_ARM64, CS_MODE_ARM)
         self.cs.detail = True
 
-        # Worker to parse dyld bytecode stream and extract dyld stub addresses to the DyldBoundSymbol they represent
-        self.dyld_info_parser = DyldInfoParser(self.binary)
         # Each __stubs function calls a single dyld stub address, which has a corresponding DyldBoundSymbol.
         # Map of each __stub function to the associated name of the DyldBoundSymbol
         self._imported_symbol_addresses_to_names: Dict[VirtualMemoryPointer, str] = {}
@@ -183,6 +182,12 @@ class MachoAnalyzer:
         self._build_callable_symbol_index()
         self._build_function_boundaries_index()
 
+        self._cfstring_to_stringref_map = self._build_cfstring_map()
+        self._cstring_to_stringref_map = self._build_cstring_map()
+
+        self.__cached_strings: Optional[Set[str]] = None
+        self.__cached_cstrings: Optional[Set[str]] = None
+
         # Done setting up, store this analyzer in class cache
         MachoAnalyzer._ANALYZER_CACHE[binary] = self
 
@@ -191,8 +196,7 @@ class MachoAnalyzer:
 
     @_requires_xrefs_computed
     def calls_to(self, address: VirtualMemoryPointer) -> List[CallerXRef]:
-        """Return the list of code-locations within the binary which branch to the provided address.
-        """
+        """Return the list of code-locations within the binary which branch to the provided address."""
         xrefs_cursor = self._db_handle.execute(
             "SELECT * from function_calls WHERE destination_address=?", (int(address),)
         )
@@ -200,7 +204,7 @@ class MachoAnalyzer:
 
     @_requires_xrefs_computed
     def objc_calls_to(
-        self, objc_class_names: List[str], objc_selectors: List[str], requires_class_and_sel_found: bool,
+        self, objc_class_names: List[str], objc_selectors: List[str], requires_class_and_sel_found: bool
     ) -> List[ObjcMsgSendXref]:
         """Return the list of code-locations in the binary which invoke _objc_msgSend with any of the provided
         classes or selectors.
@@ -232,7 +236,7 @@ class MachoAnalyzer:
         try:
             from strongarm_dataflow.dataflow import compute_function_basic_blocks_fast
         except ImportError as e:
-            if "libcapstone" in e:
+            if "libcapstone" in str(e):
                 import sys
 
                 print("\ncapstone 4.x could not be found, is the capstone backend installed?\n")
@@ -249,8 +253,7 @@ class MachoAnalyzer:
     def get_basic_block_boundaries(
         self, entry_point: VirtualMemoryPointer
     ) -> List[Tuple[VirtualMemoryPointer, VirtualMemoryPointer]]:
-        """Given the function starting at the provided address, return the list of (start_addr, end_addr) basic blocks.
-        """
+        """Given the function starting at the provided address, return the list of (start_addr, end_addr) basic blocks."""  # noqa: E501
         cursor = self._db_handle.execute(
             "SELECT start_address, end_address FROM basic_blocks WHERE entry_point=?", (entry_point,)
         )
@@ -304,7 +307,7 @@ class MachoAnalyzer:
         return VirtualMemoryPointer(objc_msgsend_symbol.address)
 
     @cached_property
-    def _objc_fastpath_ptrs_to_selector_names(self):
+    def _objc_fastpath_ptrs_to_selector_names(self) -> Dict[VirtualMemoryPointer, str]:
         # Some special selectors have a fast-path, _objc_opt_<selector>, that was added in iOS 13
         # The _objc_opt_* call is emitted instead of _objc_msgSend when the NSObject implementation will be called.
         # Found with: $ nm "/usr/lib/libobjc.A.dylib" | grep "objc_opt"
@@ -336,11 +339,11 @@ class MachoAnalyzer:
         from strongarm_dataflow.dataflow import build_xref_database_fast
 
         if self._has_computed_xrefs:
-            logging.error("Already computed xrefs, why was _build_xref_database called again?")
+            logger.error("Already computed xrefs, why was _build_xref_database called again?")
             return
 
         start_time = time.time()
-        logging.debug(f"{self.binary.path} computing call XRefs...")
+        logger.debug(f"{self.binary.path} computing call XRefs...")
 
         objc_function_family = list(self._objc_fastpath_ptrs_to_selector_names.keys())
         if self._objc_msgSend_addr:
@@ -359,7 +362,7 @@ class MachoAnalyzer:
 
         self._has_computed_xrefs = True
         end_time = time.time()
-        logging.debug(f"Finding xrefs took {end_time - start_time} seconds")
+        logger.debug(f"Finding xrefs took {end_time - start_time} seconds")
 
     @classmethod
     def clear_cache(cls) -> None:
@@ -367,7 +370,7 @@ class MachoAnalyzer:
         This can be used when you are finished analyzing a binary set and don't want to retain the cached data in memory
         """
         for binary, analyzer in cls._ANALYZER_CACHE.items():
-            logging.debug(f"Deleting db {analyzer._db_path}...")
+            logger.debug(f"Deleting db {analyzer._db_path}...")
             analyzer._db_handle.close()
             shutil.rmtree(analyzer._db_tempdir.as_posix())
 
@@ -376,13 +379,12 @@ class MachoAnalyzer:
     @property
     def objc_helper(self) -> ObjcRuntimeDataParser:
         if not self._objc_helper:
-            self._objc_helper = ObjcRuntimeDataParser(self.binary, self.dyld_info_parser)
+            self._objc_helper = ObjcRuntimeDataParser(self.binary)
         return self._objc_helper
 
     @classmethod
     def get_analyzer(cls, binary: MachoBinary) -> "MachoAnalyzer":
-        """Get a cached analyzer for a given MachoBinary
-        """
+        """Get a cached analyzer for a given MachoBinary."""
         if binary in cls._ANALYZER_CACHE:
             # There exists a MachoAnalyzer for this binary - use it instead of making a new one
             return cls._ANALYZER_CACHE[binary]
@@ -399,32 +401,27 @@ class MachoAnalyzer:
         return None
 
     def objc_classes(self) -> List[ObjcClass]:
-        """Return the List of classes and categories implemented within the binary
-        """
+        """Return the List of classes and categories implemented within the binary."""
         return self.objc_helper.classes
 
     def objc_categories(self) -> List[ObjcCategory]:
-        """Return the List of categories implemented within the app
-        """
+        """Return the List of categories implemented within the app."""
         all_classes = self.objc_classes()
         categories: List[ObjcCategory] = [c for c in all_classes if isinstance(c, ObjcCategory)]
         return categories
 
     def get_conformed_protocols(self) -> List[ObjcProtocol]:
-        """Return the List of protocols to which code within the binary conforms
-        """
+        """Return the List of protocols to which code within the binary conforms."""
         return self.objc_helper.protocols
 
     @property
     def dyld_bound_symbols(self) -> Dict[VirtualMemoryPointer, DyldBoundSymbol]:
-        """Return a Dict of each imported dyld stub to the corresponding symbol to be bound at runtime.
-        """
-        return self.dyld_info_parser.dyld_stubs_to_symbols
+        """Return a Dict of each imported dyld stub to the corresponding symbol to be bound at runtime."""
+        return self.binary.dyld_bound_symbols
 
     @property
     def imp_stubs_to_symbol_names(self) -> Dict[VirtualMemoryPointer, str]:
-        """Return a Dict of callable implementation stubs to the names of the imported symbols they correspond to.
-        """
+        """Return a Dict of callable implementation stubs to the names of the imported symbols they correspond to."""
         if self._imported_symbol_addresses_to_names:
             return self._imported_symbol_addresses_to_names
 
@@ -483,22 +480,19 @@ class MachoAnalyzer:
         return {y: x for x, y in self.exported_symbol_pointers_to_names.items()}
 
     def exported_symbol_name_for_address(self, address: VirtualMemoryPointer) -> Optional[str]:
-        """Return the symbol name for the provided address, or None if the address is not a named exported symbol.
-        """
+        """Return the symbol name for the provided address, or None if the address is not a named exported symbol."""
         if address in self.exported_symbol_pointers_to_names:
             return self.exported_symbol_pointers_to_names[address]
         return None
 
     def symbol_name_for_branch_destination(self, branch_address: VirtualMemoryPointer) -> str:
-        """Get the associated symbol name for a given branch destination
-        """
+        """Get the associated symbol name for a given branch destination."""
         if branch_address in self.imp_stubs_to_symbol_names:
             return self.imp_stubs_to_symbol_names[branch_address]
         raise RuntimeError(f"Unknown branch destination {hex(branch_address)}. Is this a local branch?")
 
     def disassemble_region(self, start_address: VirtualMemoryPointer, size: int) -> List[CsInsn]:
-        """Disassemble the executable code in a given region into a list of CsInsn objects
-        """
+        """Disassemble the executable code in a given region into a list of CsInsn objects."""
         func_str = bytes(self.binary.get_content_from_virtual_address(virtual_address=start_address, size=size))
         instructions = [instr for instr in self.cs.disasm(func_str, start_address)]
         if not len(instructions):
@@ -506,8 +500,7 @@ class MachoAnalyzer:
         return instructions
 
     def get_function_instructions(self, start_address: VirtualMemoryPointer) -> List[CsInsn]:
-        """Get a list of disassembled instructions for the function beginning at start_address
-        """
+        """Get a list of disassembled instructions for the function beginning at start_address."""
         end_address = self.get_function_end_address(start_address)
 
         if end_address is None:
@@ -529,8 +522,7 @@ class MachoAnalyzer:
         return self.objc_helper.selector_for_selector_literal(selref_ptr)
 
     def get_method_imp_addresses(self, selector: str) -> List[VirtualMemoryPointer]:
-        """Given a selector, return a list of virtual addresses corresponding to the start of each IMP for that SEL
-        """
+        """Given a selector, return a list of virtual addresses corresponding to the start of each IMP for that SEL."""
         return self.objc_helper.get_method_imp_addresses(selector)
 
     def get_imps_for_sel(self, selector: str) -> List["ObjcFunctionAnalyzer"]:
@@ -552,8 +544,7 @@ class MachoAnalyzer:
         return implementation_analyzers
 
     def get_objc_methods(self) -> List["ObjcMethodInfo"]:
-        """Get a List of ObjcMethodInfo's representing all ObjC methods implemented in the Mach-O.
-        """
+        """Get a List of ObjcMethodInfo's representing all ObjC methods implemented in the Mach-O."""
         from strongarm.objc import ObjcMethodInfo  # type: ignore
 
         if self._objc_method_list:
@@ -594,13 +585,18 @@ class MachoAnalyzer:
 
         return VirtualMemoryPointer(results[0])
 
+    @cached_property
+    def class_for_class_pointer_map(self) -> Dict[VirtualMemoryPointer, ObjcClass]:
+        return {VirtualMemoryPointer(x.raw_struct.binary_offset): x for x in self.objc_classes()}
+
     def class_name_for_class_pointer(self, classref: VirtualMemoryPointer) -> Optional[str]:
         """Given a classref, return the name of the class.
         This method will handle classes implemented within the binary and imported classes.
         """
         # Did the caller provide a classref for an imported class?
-        if classref in self.imported_symbols_to_symbol_names:
-            return self.imported_symbols_to_symbol_names[classref]
+        local_class = self.imported_symbols_to_symbol_names.get(classref)
+        if local_class:
+            return local_class
 
         # The class is implemented within the binary and has an associated ObjcClass object
         # We could have been passed either a classref pointer in __objc_classrefs, or the direct address of
@@ -608,10 +604,9 @@ class MachoAnalyzer:
 
         # First, check if we were provided with the address of an __objc_data struct in __objc_data representing
         # the class.
-        local_class = [x for x in self.objc_classes() if x.raw_struct.binary_offset == classref]
-        if len(local_class):
-            assert len(local_class) == 1
-            return local_class[0].name
+        local_class = self.class_for_class_pointer_map.get(classref)
+        if local_class:
+            return local_class.name
 
         # Then, check if we were passed a classref pointer in __objc_classrefs
         try:
@@ -620,17 +615,15 @@ class MachoAnalyzer:
             # Invalid classref
             return None
 
-        local_class = [x for x in self.objc_classes() if x.raw_struct.binary_offset == dereferenced_classref]
-        if len(local_class):
-            assert len(local_class) == 1
-            return local_class[0].name
+        local_class = self.class_for_class_pointer_map.get(dereferenced_classref)
+        if local_class:
+            return local_class.name
 
         # Invalid classref
         return None
 
     def classref_for_class_name(self, class_name: str) -> Optional[VirtualMemoryPointer]:
-        """Given a class name, try to find a classref for it.
-        """
+        """Given a class name, try to find a classref for it."""
         classrefs = [
             addr
             for addr, name in self.imported_symbols_to_symbol_names.items()
@@ -640,8 +633,6 @@ class MachoAnalyzer:
             return classrefs[0]
 
         # TODO(PT): this is expensive! We should do one analysis step of __objc_classrefs and create a map.
-        classref_locations, classref_destinations = self.binary.read_pointer_section("__objc_classrefs")
-
         # is it a local class?
         class_locations = [x.raw_struct.binary_offset for x in self.objc_classes() if x.name == class_name]
         if not len(class_locations):
@@ -649,92 +640,92 @@ class MachoAnalyzer:
             return None
         class_location = VirtualMemoryPointer(class_locations[0])
 
-        if class_location not in classref_destinations:
-            # unknown class name
-            return None
-
-        classref_index = classref_destinations.index(class_location)
-        return classref_locations[classref_index]
+        classref_addr_to_pointer_map = self.binary.read_pointer_section("__objc_classrefs")
+        # If None is returned, it is an unknown class name
+        return first((k for k, v in classref_addr_to_pointer_map.items() if v == class_location), None)
 
     def selref_for_selector_name(self, selector_name: str) -> Optional[VirtualMemoryPointer]:
         return self.objc_helper.selref_for_selector_name(selector_name)
 
     @_requires_xrefs_computed
     def strings(self) -> Set[str]:
-        """Return a list containing every string in the binary
-        """
-        # Gather strings from various sections
-        all_strings = set()
-        for section_name in ["__cstring", "__objc_methname", "__objc_methtype", "__objc_classname"]:
-            section_strings = self._strings_in_section(section_name)
-            all_strings.update(section_strings)
+        """Return a list containing every string in the binary."""
+        if not self.__cached_strings:
+            # Gather strings from various sections
+            all_strings = set()
+            all_strings.update(self.get_cstrings())
+            for section_name in ["__objc_methname", "__objc_methtype", "__objc_classname", "__const"]:
+                section_strings = self._strings_in_section(section_name)
+                all_strings.update(section_strings)
 
-        # Gather strings found via Xrefs
-        c = self._db_handle.cursor()
-        xref_strings_rows = c.execute("SELECT string_literal from string_xrefs").fetchall()
-        xref_strings = [xref_strings_row[0] for xref_strings_row in xref_strings_rows]
-        all_strings.update(set(xref_strings))
+            # Gather strings found via Xrefs
+            c = self._db_handle.cursor()
+            xref_strings_rows = c.execute("SELECT string_literal from string_xrefs").fetchall()
+            xref_strings = [xref_strings_row[0] for xref_strings_row in xref_strings_rows]
+            all_strings.update(set(xref_strings))
 
-        return all_strings
+            self.__cached_strings = all_strings
+        return self.__cached_strings
 
     def get_cstrings(self) -> Set[str]:
-        """Return the list of strings in the binary's __cstring section.
-        """
-        # TODO(PT): This is SLOW and WASTEFUL!!!
-        # These transformations should be done ONCE on initial analysis!
-        # This method should cache its result.
-        return self._strings_in_section("__cstring")
+        """Return the list of strings in the binary's __cstring section."""
+        if not self.__cached_cstrings:
+            self.__cached_cstrings = self._strings_in_section("__cstring")
+        return self.__cached_cstrings
+
+    def _build_cstring_map(self) -> Dict[str, VirtualMemoryPointer]:
+        strings_section = self.binary.section_with_name("__cstring", "__TEXT")
+        if not strings_section:
+            return {}
+
+        strings_base = strings_section.address
+        strings_content = self.binary.get_bytes(strings_section.offset, strings_section.size)
+
+        string_to_stringrefs = {}
+        # split into characters (string table is packed and each entry is terminated by a null character)
+        string_table = list(strings_content)
+        transformed_strings = MachoStringTableHelper.transform_string_section(string_table)
+        for idx, entry in transformed_strings.items():
+            # Address is the base of __cstring plus the index of the entry
+            stringref_address = VirtualMemoryPointer(strings_base + idx)
+            string_to_stringrefs[entry.full_string] = stringref_address
+        return string_to_stringrefs
 
     def _stringref_for_cstring(self, string: str) -> Optional[VirtualMemoryPointer]:
         """Try to find the stringref in __cstrings for a provided C string.
         If the string is not present in the __cstrings section, this method returns None.
         """
-        # TODO(PT): This is SLOW and WASTEFUL!!!
-        # These transformations should be done ONCE on initial analysis!
-        strings_section = self.binary.section_with_name("__cstring", "__TEXT")
-        if not strings_section:
+        if string not in self._cstring_to_stringref_map:
             return None
+        return self._cstring_to_stringref_map[string]
 
-        strings_base = strings_section.address
-        strings_content = strings_section.content
+    def _build_cfstring_map(self) -> Dict[str, VirtualMemoryPointer]:
+        cfstrings_section = self.binary.section_with_name("__cfstring", "__DATA")
+        if not cfstrings_section:
+            cfstrings_section = self.binary.section_with_name("__cfstring", "__DATA_CONST")
+            if not cfstrings_section:
+                return {}
 
-        # split into characters (string table is packed and each entry is terminated by a null character)
-        string_table = list(strings_content)
-        transformed_strings = MachoStringTableHelper.transform_string_section(string_table)
-        for idx, entry in transformed_strings.items():
-            if entry.full_string == string:
-                # found the string we're looking for
-                # the address is the base of __cstring plus the index of the entry
-                stringref_address = strings_base + idx
-                return stringref_address
+        sizeof_cfstring = sizeof(CFString64) if self.binary.is_64bit else sizeof(CFString32)
+        cfstrings_base = cfstrings_section.address
 
-        # didn't find the string the user requested
-        return None
+        cfstring_to_stringrefs = {}
+        cfstrings_count = int((cfstrings_section.end_address - cfstrings_section.address) / sizeof_cfstring)
+        for i in range(cfstrings_count):
+            cfstring_addr = cfstrings_base + (i * sizeof_cfstring)
+            cfstring = self.binary.read_struct_with_rebased_pointers(cfstring_addr, CFStringStruct, virtual=True)
+            literal = self.binary.read_string_at_address(cfstring.literal)
+            if literal:
+                cfstring_to_stringrefs[literal] = VirtualMemoryPointer(cfstring_addr)
+        return cfstring_to_stringrefs
 
     def _stringref_for_cfstring(self, string: str) -> Optional[VirtualMemoryPointer]:
         """Try to find the stringref in __cfstrings for a provided Objective-C string literal.
         If the string is not present in the __cfstrings section, this method returns None.
         """
-        # TODO(PT): This is SLOW and WASTEFUL!!!
-        # These transformations should be done ONCE on initial analysis!
-        cfstrings_section = self.binary.section_with_name("__cfstring", "__DATA")
-        if not cfstrings_section:
+        if string not in self._cfstring_to_stringref_map:
             return None
-
-        sizeof_cfstring = sizeof(CFString64) if self.binary.is_64bit else sizeof(CFString32)
-        cfstrings_base = cfstrings_section.address
-
-        cfstrings_count = int((cfstrings_section.end_address - cfstrings_section.address) / sizeof_cfstring)
-        for i in range(cfstrings_count):
-            cfstring_addr = cfstrings_base + (i * sizeof_cfstring)
-            cfstring = self.binary.read_struct(cfstring_addr, CFStringStruct, virtual=True)
-
-            # check if this is the string the user requested
-            string_address = cfstring.literal
-            if self.binary.read_string_at_address(string_address) == string:
-                return VirtualMemoryPointer(cfstring_addr)
-
-        return None
+        return self._cfstring_to_stringref_map[string]
 
     def stringref_for_string(self, string: str) -> Optional[VirtualMemoryPointer]:
         """Try to find the stringref for a provided string.
@@ -834,11 +825,11 @@ class MachoAnalyzer:
         self._db_handle.commit()
 
     def _strings_in_section(self, section_name: str) -> Set[str]:
-        """Fetch the list of strings located inside the provided section
-        """
+        """Fetch the list of strings located inside the provided section."""
         discovered_strings = set()
         string_section = self.binary.section_with_name(section_name, "__TEXT")
         if string_section:
-            transformed_strings = MachoStringTableHelper.transform_string_section(list(string_section.content))
+            strings_content = self.binary.get_bytes(string_section.offset, string_section.size)
+            transformed_strings = MachoStringTableHelper.transform_string_section(list(strings_content))
             discovered_strings = set((x.full_string for x in transformed_strings.values()))
         return discovered_strings
