@@ -2,11 +2,11 @@ import math
 from ctypes import c_uint32, c_uint64, sizeof
 from distutils.version import LooseVersion
 from pathlib import Path
-from typing import TYPE_CHECKING, Any, List, Optional, Set, Tuple, Type, TypeVar
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Type, TypeVar
 
 from _ctypes import Structure
 
-from strongarm.debug_util import DebugUtil
+from strongarm.logger import strongarm_logger
 from strongarm.macho.arch_independent_structs import (
     ArchIndependentStructure,
     CFStringStruct,
@@ -41,34 +41,31 @@ from strongarm.macho.macho_load_commands import MachoLoadCommands
 if TYPE_CHECKING:
     from strongarm.macho.codesign import CodesignParser
 
+logger = strongarm_logger.getChild(__file__)
+
 AIS = TypeVar("AIS", bound=ArchIndependentStructure)
 
 
 class BinaryEncryptedError(Exception):
-    """Raised when the binary is encrypted.
-    """
+    """Raised when the binary is encrypted."""
 
 
 class LoadCommandMissingError(Exception):
-    """Raised when the binary is missing a load command.
-    """
+    """Raised when the binary is missing a load command."""
 
 
 class NoEmptySpaceForLoadCommandError(Exception):
-    """Raised when we fail to insert a load command because there's not enough empty space left in the Mach-O header.
-    """
+    """Raised when we fail to insert a load command because there's not enough empty space left in the Mach-O header."""
 
 
 class InvalidAddressError(Exception):
-    """Raised when a client asks for bytes at an address outside the binary.
-    """
+    """Raised when a client asks for bytes at an address outside the binary."""
 
 
 class MachoSegment:
-    def __init__(self, binary: "MachoBinary", segment_command: MachoSegmentCommandStruct) -> None:
+    def __init__(self, segment_command: MachoSegmentCommandStruct) -> None:
         self.cmd = segment_command
 
-        self.content = binary.get_bytes(segment_command.fileoff, segment_command.filesize)
         self.name = segment_command.segname.decode()
         self.sizeof = segment_command.sizeof
 
@@ -94,12 +91,11 @@ class MachoSegment:
 
 
 class MachoSection:
-    def __init__(self, binary: "MachoBinary", section_command: MachoSectionRawStruct, segment: MachoSegment) -> None:
+    def __init__(self, section_command: MachoSectionRawStruct, segment: MachoSegment) -> None:
         self.cmd = section_command
         self.segment = segment
 
         # ignore these types due to dynamic attributes of associated types
-        self.content = binary.get_bytes(section_command.offset, section_command.size)
         self.name = section_command.sectname.decode()
         self.segment_name = section_command.segname.decode()
         self.address = section_command.addr
@@ -125,8 +121,7 @@ class MachoBinary:
     BYTES_PER_INSTRUCTION = 4
 
     def __init__(self, path: Path, binary_data: bytes, file_offset: Optional[StaticFilePointer] = None) -> None:
-        """Parse the bytes representing a Mach-O file.
-        """
+        """Parse the bytes representing a Mach-O file."""
         from .codesign.codesign_parser import CodesignParser
 
         self._cached_binary = binary_data
@@ -153,6 +148,8 @@ class MachoBinary:
         self._symtab: Optional[MachoSymtabCommandStruct] = None
         self._encryption_info: Optional[MachoEncryptionInfoStruct] = None
         self._dyld_info: Optional[MachoDyldInfoCommandStruct] = None
+        self._dyld_export_trie: Optional[MachoLinkeditDataCommandStruct] = None
+        self._dyld_chained_fixups: Optional[MachoLinkeditDataCommandStruct] = None
         self.load_dylib_commands: List[DylibCommandStruct] = []
         self._code_signature_cmd: Optional[MachoLinkeditDataCommandStruct] = None
         self._function_starts_cmd: Optional[MachoLinkeditDataCommandStruct] = None
@@ -162,14 +159,27 @@ class MachoBinary:
         self._build_tool_versions: Optional[List[MachoBuildToolVersionStruct]] = None
 
         self.__codesign_parser: Optional[CodesignParser] = None
+        self.__minimum_deployment_target: Optional[LooseVersion] = None
 
         # This kicks off the parse of the binary
         if not self.parse():
             raise RuntimeError("Failed to parse Mach-O")
 
         self.platform_word_type = c_uint64 if self.is_64bit else c_uint32
-        self.symtab_contents = self._get_symtab_contents()
-        DebugUtil.log(self, f"parsed symtab, len = {len(self.symtab_contents)}")
+
+        self._symtab_contents: Optional[List[MachoNlistStruct]] = None
+        logger.debug(self, f"parsed symtab, len = {len(self.symtab_contents)}")
+
+        from .dyld_info_parser import DyldBoundSymbol, DyldInfoParser
+
+        self.dyld_bound_symbols: Dict[VirtualMemoryPointer, DyldBoundSymbol] = {}
+        self.dyld_rebased_pointers: Dict[VirtualMemoryPointer, VirtualMemoryPointer] = {}
+
+        if self._dyld_chained_fixups:
+            rebases, binds = DyldInfoParser.parse_chained_fixups(self)  # type: ignore
+            self.dyld_rebased_pointers, self.dyld_bound_symbols = rebases, binds
+        else:
+            self.dyld_bound_symbols = DyldInfoParser.parse_dyld_info(self)
 
     def __repr__(self) -> str:
         return f"<MachoBinary binary={self.path}>"
@@ -182,11 +192,11 @@ class MachoBinary:
             False otherwise.
 
         """
-        # DebugUtil.log(self, f"parsing Mach-O slice @ {hex(int(self._offset_within_fat))} in {self.path}")
+        # logger.debug(self, f"parsing Mach-O slice @ {hex(int(self._offset_within_fat))} in {self.path}")
 
         # Preliminary Mach-O parsing
         if not self.verify_magic():
-            DebugUtil.log(self, f"unsupported magic {hex(self.slice_magic)}")
+            logger.debug(self, f"unsupported magic {hex(self.slice_magic)}")
             return False
 
         self.is_swap = self.should_swap_bytes()
@@ -198,13 +208,12 @@ class MachoBinary:
 
         self.parse_header()
 
-        DebugUtil.log(self, f"header parsed. non-native endianness? {self.is_swap}. 64-bit? {self.is_64bit}")
+        logger.debug(self, f"header parsed. non-native endianness? {self.is_swap}. 64-bit? {self.is_64bit}")
         return True
 
     @property
     def slice_magic(self) -> int:
-        """Read magic number identifier from this Mach-O slice
-        """
+        """Read magic number identifier from this Mach-O slice."""
         return self.read_word(0, virtual=False, word_type=c_uint32)
 
     def verify_magic(self) -> bool:
@@ -249,8 +258,7 @@ class MachoBinary:
         self._parse_load_commands(load_commands_off, self.header.ncmds)  # type: ignore
 
     def _parse_header_flags(self) -> None:
-        """Interpret binary's header bitset and populate self.header_flags
-        """
+        """Interpret binary's header bitset and populate self.header_flags."""
         self.header_flags = []
 
         flags_bitset = self.header.flags
@@ -275,7 +283,7 @@ class MachoBinary:
             if load_command.cmd in [MachoLoadCommands.LC_SEGMENT, MachoLoadCommands.LC_SEGMENT_64]:
                 segment_command = self.read_struct(offset, MachoSegmentCommandStruct)
                 # TODO(PT) handle byte swap of segment if necessary
-                segment = MachoSegment(self, segment_command)
+                segment = MachoSegment(segment_command)
                 self.segments.append(segment)
                 self._parse_sections_for_segment(segment, offset)
 
@@ -293,6 +301,12 @@ class MachoBinary:
             elif load_command.cmd in [MachoLoadCommands.LC_DYLD_INFO, MachoLoadCommands.LC_DYLD_INFO_ONLY]:
                 self._dyld_info = self.read_struct(offset, MachoDyldInfoCommandStruct)
 
+            elif load_command.cmd in [MachoLoadCommands.LC_DYLD_EXPORTS_TRIE]:
+                self._dyld_export_trie = self.read_struct(offset, MachoLinkeditDataCommandStruct)
+
+            elif load_command.cmd in [MachoLoadCommands.LC_DYLD_CHAINED_FIXUPS]:
+                self._dyld_chained_fixups = self.read_struct(offset, MachoLinkeditDataCommandStruct)
+
             elif load_command.cmd in [MachoLoadCommands.LC_LOAD_DYLIB, MachoLoadCommands.LC_LOAD_WEAK_DYLIB]:
                 dylib_load_command = self.read_struct(offset, DylibCommandStruct)
                 self.load_dylib_commands.append(dylib_load_command)
@@ -334,17 +348,43 @@ class MachoBinary:
         data = self.get_contents_from_address(address=binary_offset, size=sizeof(backing_layout), is_virtual=virtual)
         return struct_type(binary_offset, data, backing_layout)
 
-    def section_name_for_address(self, virt_addr: VirtualMemoryPointer) -> Optional[str]:
-        """Given an address in the virtual address space, return the name of the section which contains it.
+    def read_struct_with_rebased_pointers(
+        self, binary_offset: int, struct_type: Type[AIS], virtual: bool = False
+    ) -> AIS:
+        """Read a static binary structure that may contain rebased pointers.
+        For each uint64_t within the structure, check if we know that this pointer should be rebased.
+        If so, the static data here may be a packed chained fixup pointer, rather than a pointer we can follow.
+        In this case, update the pointer to contain the value to be rebased, so that the pointer can be followed.
         """
+        backing_layout = struct_type.get_backing_data_layout(self.is_64bit, self.get_minimum_deployment_target())
+        data = self.get_contents_from_address(address=binary_offset, size=sizeof(backing_layout), is_virtual=virtual)
+        s = struct_type(binary_offset, data, backing_layout)
+
+        # Check each c_uint64/c_ulong to see if we know about a rebase for it. If so, apply it
+        base_virt_offset = binary_offset
+        if not virtual:
+            base_virt_offset += self.get_virtual_base()
+        for field_name, field_type, *_ in backing_layout._fields_:
+            field_offset = getattr(getattr(backing_layout, field_name), "offset")
+            field_address = base_virt_offset + field_offset
+            if field_type == c_uint64 and field_address in self.dyld_rebased_pointers:
+                logger.debug(
+                    f"Setting rebased pointer within {struct_type}+{field_offset} -> "
+                    f"{self.dyld_rebased_pointers[field_address]} at {field_address}"
+                )
+                setattr(s, field_name, self.dyld_rebased_pointers[field_address])
+
+        return s
+
+    def section_name_for_address(self, virt_addr: VirtualMemoryPointer) -> Optional[str]:
+        """Given an address in the virtual address space, return the name of the section which contains it."""
         section = self.section_for_address(virt_addr)
         if not section:
             return None
         return section.name
 
     def section_for_address(self, virt_addr: VirtualMemoryPointer) -> Optional[MachoSection]:
-        """Given an address in the virtual address space, return the section which contains it.
-        """
+        """Given an address in the virtual address space, return the section which contains it."""
         # invalid address?
         if virt_addr < self.get_virtual_base():
             return None
@@ -372,8 +412,7 @@ class MachoBinary:
             raise ValueError(f"segment_index ({segment_index}) out of bounds ({len(self.segments)}")
 
     def segment_with_name(self, desired_segment_name: str) -> Optional[MachoSegment]:
-        """Returns the segment with the provided name. Returns None if there's no such segment in the binary.
-        """
+        """Returns the segment with the provided name. Returns None if there's no such segment in the binary."""
         # TODO(PT): add unit test for this method
         return next((s for s in self.segments if s.name == desired_segment_name), None)
 
@@ -403,7 +442,7 @@ class MachoBinary:
             # TODO(PT): handle byte swap of segment
             section_command = self.read_struct(section_offset, MachoSectionRawStruct)
             # Encapsulate header and content into one object, and store that
-            section = MachoSection(self, section_command, segment)
+            section = MachoSection(section_command, segment)
             segment.sections.append(section)
             # Add to list of sections within the Mach-O
             self.sections.append(section)
@@ -427,8 +466,7 @@ class MachoBinary:
         return self._virtual_base
 
     def get_file_offset(self) -> StaticFilePointer:
-        """Retrieve the offset within the file of this Mach-O slice
-        """
+        """Retrieve the offset within the file of this Mach-O slice."""
         return self.file_offset
 
     def get_bytes(self, offset: StaticFilePointer, size: int, _translate_addr_to_file: bool = False) -> bytearray:
@@ -457,10 +495,10 @@ class MachoBinary:
 
         # safeguard against reading from an encrypted segment of the binary
         if self.is_range_encrypted(offset, size):
+            encryption_range_start = int(self.encryption_info.cryptoff)
+            encryption_range_end = encryption_range_start + int(self.encryption_info.cryptsize)
             raise BinaryEncryptedError(
-                "Cannot read encrypted"
-                f" range [{hex(int(self.encryption_info.cryptoff))}"
-                f" to {hex(int(self.encryption_info.cryptsize))}]"
+                f"Cannot read encrypted range [{hex(encryption_range_start)} - {hex(encryption_range_end)}]"
             )
 
         return bytearray(self._cached_binary[offset : offset + size])
@@ -487,14 +525,20 @@ class MachoBinary:
         string_table = list(string_table_data)
         return string_table
 
-    def _get_symtab_contents(self) -> List[MachoNlistStruct]:
+    @property
+    def symtab_contents(self) -> List[MachoNlistStruct]:
+        if self._symtab_contents is None:
+            self._symtab_contents = self._parse_symtab_contents()
+            logger.debug(self, f"parsed symtab, len = {len(self.symtab_contents)}")
+        return self._symtab_contents
+
+    def _parse_symtab_contents(self) -> List[MachoNlistStruct]:
         """Parse symbol table containing list of Nlist64's
 
         Returns:
             Array of Nlist64's representing binary's symbol table
-
         """
-        DebugUtil.log(self, f"parsing {self.symtab.nsyms} symtab entries")
+        logger.debug(self, f"parsing {self.symtab.nsyms} symtab entries")
 
         symtab = []
         # start reading from symoff and increment by one Nlist64 each iteration
@@ -611,15 +655,13 @@ class MachoBinary:
         return self.get_full_string_from_start_address(address)
 
     def is_encrypted(self) -> bool:
-        """Returns True if the binary has an encrypted segment, False otherwise
-        """
+        """Returns True if the binary has an encrypted segment, False otherwise."""
         if not self._encryption_info:
             return False
         return self.encryption_info.cryptid != 0
 
     def is_range_encrypted(self, offset: StaticFilePointer, size: int) -> bool:
-        """Returns whether the provided address range overlaps with the encrypted section of the binary.
-        """
+        """Returns whether the provided address range overlaps with the encrypted section of the binary."""
         if not self.is_encrypted():
             return False
 
@@ -643,8 +685,7 @@ class MachoBinary:
         return self.load_dylib_commands[idx]
 
     def dylib_name_for_library_ordinal(self, library_ordinal: int) -> str:
-        """Read the name of the dynamic library by its library ordinal
-        """
+        """Read the name of the dynamic library by its library ordinal."""
         source_dylib = self.dylib_for_library_ordinal(library_ordinal)
         if source_dylib:
             source_name_addr = source_dylib.binary_offset + source_dylib.dylib.name.offset + self.get_virtual_base()
@@ -659,7 +700,7 @@ class MachoBinary:
             source_name = "<unknown dylib>"
         return source_name
 
-    def read_pointer_section(self, section_name: str) -> Tuple[List[VirtualMemoryPointer], List[VirtualMemoryPointer]]:
+    def read_pointer_section(self, section_name: str) -> Dict[VirtualMemoryPointer, VirtualMemoryPointer]:
         """Read all the pointers in a section
 
         It is the caller's responsibility to only call this with a `section_name` which indicates a section which should
@@ -672,9 +713,6 @@ class MachoBinary:
         The indexes of these two lists are matched up; that is, list1[0] is the virtual address of the first pointer
         in the requested section, and list2[0] is the pointer value contained at that address.
         """
-        locations: List[VirtualMemoryPointer] = []
-        entries: List[VirtualMemoryPointer] = []
-
         # PT: Assume a pointer-list-section will always be in __DATA or __DATA_CONST. True as far as I know.
         for segment in ["__DATA", "__DATA_CONST"]:
             section = self.section_with_name(section_name, segment)
@@ -682,10 +720,12 @@ class MachoBinary:
                 break
         else:
             # Couldn't find the desired section
-            return locations, entries
+            return {}
+
+        address_to_pointer_map: Dict[VirtualMemoryPointer, VirtualMemoryPointer] = {}
 
         section_base = section.address
-        section_data = section.content
+        section_data = self.get_bytes(section.offset, section.size)
 
         binary_word = self.platform_word_type
         pointer_count = int(len(section_data) / sizeof(binary_word))
@@ -693,19 +733,26 @@ class MachoBinary:
 
         for i in range(pointer_count):
             # convert section offset of entry to absolute virtual address
-            locations.append(VirtualMemoryPointer(section_base + pointer_off))
+            ptr_location = VirtualMemoryPointer(section_base + pointer_off)
 
-            data_end = pointer_off + sizeof(binary_word)
-            val = binary_word.from_buffer(bytearray(section_data[pointer_off:data_end])).value
-            entries.append(VirtualMemoryPointer(val))
+            if ptr_location in self.dyld_rebased_pointers:
+                ptr_value = self.dyld_rebased_pointers[ptr_location]
+                logger.debug(f"Pointer is rebased: {ptr_location} -> {ptr_value}")
+            else:
+                data_end = pointer_off + sizeof(binary_word)
+                ptr_value = VirtualMemoryPointer(
+                    binary_word.from_buffer(bytearray(section_data[pointer_off:data_end])).value
+                )
+                logger.debug(f"Pointer was not in the rebase list: {ptr_location} -> {ptr_value}")
+
+            address_to_pointer_map[ptr_location] = VirtualMemoryPointer(ptr_value)
 
             pointer_off += sizeof(binary_word)
 
-        return locations, entries
+        return address_to_pointer_map
 
     def read_word(self, address: int, virtual: bool = True, word_type: Any = None) -> int:
-        """Attempt to read a word from the binary at a virtual address.
-        """
+        """Attempt to read a word from the binary at a virtual address."""
         if not word_type:
             word_type = self.platform_word_type
 
@@ -719,6 +766,16 @@ class MachoBinary:
 
         return word_type.from_buffer(bytearray(file_bytes)).value
 
+    def read_rebased_pointer(self, address: VirtualMemoryPointer) -> VirtualMemoryPointer:
+        """Attempt to read a rebased pointer from the binary at a virtual address.
+        The pointer is assumed to be the platform word size.
+        """
+        if address not in self.dyld_rebased_pointers:
+            # This may be a pre-iOS 15 binary for which we don't record rebases
+            return VirtualMemoryPointer(self.read_word(address, virtual=True, word_type=self.platform_word_type))
+
+        return self.dyld_rebased_pointers[address]
+
     @property
     def header(self) -> MachoHeaderStruct:
         if self._header:
@@ -767,18 +824,15 @@ class MachoBinary:
         return self.__codesign_parser
 
     def get_entitlements(self) -> Optional[bytearray]:
-        """Read the entitlements the binary was signed with.
-        """
+        """Read the entitlements the binary was signed with."""
         return self._codesign_parser.entitlements
 
     def get_signing_identity(self) -> Optional[str]:
-        """Read the bundle ID the binary was signed as.
-        """
+        """Read the bundle ID the binary was signed as."""
         return self._codesign_parser.signing_identifier
 
     def get_team_id(self) -> Optional[str]:
-        """Read the team ID the binary was signed with.
-        """
+        """Read the team ID the binary was signed with."""
         return self._codesign_parser.signing_team_id
 
     def write_bytes(self, data: bytes, address: int, virtual: bool = False) -> "MachoBinary":
@@ -803,6 +857,7 @@ class MachoBinary:
     def write_struct(self, struct: Structure, address: int, virtual: bool = False) -> "MachoBinary":
         """Serialize and write the provided structure the Mach-O slice, returning a new modified binary.
         Note: This will invalidate the binary's code signature, if present.
+        TODO(PT): Deprecate and move to MachoBinaryWriter
         """
         # Write the structure bytes to the binary
         # TODO(PT): byte order?
@@ -814,6 +869,7 @@ class MachoBinary:
         including the pathname.
         Raises NoEmptySpaceForLoadCommandError() if there's not enough space in the Mach-O header to add a new command.
         Note: This will invalidate the binary's code signature, if present.
+        TODO(PT): Deprecate and move to MachoBinaryWriter
         """
         if not self.is_64bit:
             raise RuntimeError("Inserting load commands is only support on 64-bit binaries")
@@ -879,6 +935,7 @@ class MachoBinary:
 
     def write_binary(self, path: Path) -> None:
         """Write the in-memory Mach-O slice to the provided path.
+        TODO(PT): Deprecate and move to MachoBinaryWriter
         """
         # Pass 'x' so the call will throw an exception if the path already exists
         with open(path, "xb") as out_file:
@@ -887,6 +944,7 @@ class MachoBinary:
     @staticmethod
     def write_fat(slices: List["MachoBinary"], path: Path) -> None:
         """Write a list of Mach-O slices into a FAT file at the provided path.
+        TODO(PT): Deprecate and move to MachoBinaryWriter
         """
         from strongarm.macho.macho_definitions import MachArch, MachoFatArch, MachoFatHeader
 
@@ -976,9 +1034,22 @@ class MachoBinary:
         self._functions_list = functions_list
         return self._functions_list
 
-    def dylib_id(self) -> Optional[str]:
-        """If the binary contains an LC_ID_DYLIB load command, return the pathname which the binary represents.
+    def get_constructor_functions(self) -> List[VirtualMemoryPointer]:
+        """Get a list of the function entry points defined in __mod_init_func. This includes C constructors.
+
+        Returns: A list of VirtualMemoryPointers corresponding to each function's entry point.
         """
+        return list(self.read_pointer_section("__mod_init_func").values())
+
+    def get_destructor_functions(self) -> List[VirtualMemoryPointer]:
+        """Get a list of the function entry points defined in __mod_term_func. This includes C destructors.
+
+        Returns: A list of VirtualMemoryPointers corresponding to each function's entry point.
+        """
+        return list(self.read_pointer_section("__mod_term_func").values())
+
+    def dylib_id(self) -> Optional[str]:
+        """If the binary contains an LC_ID_DYLIB load command, return the pathname which the binary represents."""
         if not self._id_dylib_cmd:
             return None
 
@@ -991,14 +1062,15 @@ class MachoBinary:
         return dylib_name
 
     def get_minimum_deployment_target(self) -> Optional[LooseVersion]:
-        if not self._build_version_cmd:
-            return None
-        # X.Y.Z is encoded in nibbles xxxx.yy.zz
-        encoded_min_target = self._build_version_cmd.minos
-        patch = (encoded_min_target >> (8 * 0)) & 0xFF
-        minor = (encoded_min_target >> (8 * 1)) & 0xFF
-        major = (encoded_min_target >> (8 * 2)) & 0xFFFF
-        return LooseVersion(f"{major}.{minor}.{patch}")
+        if not self.__minimum_deployment_target:
+            if self._build_version_cmd:
+                # X.Y.Z is encoded in nibbles xxxx.yy.zz
+                encoded_min_target = self._build_version_cmd.minos
+                patch = (encoded_min_target >> (8 * 0)) & 0xFF
+                minor = (encoded_min_target >> (8 * 1)) & 0xFF
+                major = (encoded_min_target >> (8 * 2)) & 0xFFFF
+                self.__minimum_deployment_target = LooseVersion(f"{major}.{minor}.{patch}")
+        return self.__minimum_deployment_target
 
     def get_build_version_platform(self) -> Optional[MachoBuildVersionPlatform]:
         if not self._build_version_cmd:
