@@ -30,8 +30,7 @@ class _BasePointer(int):
 
 
 class StaticFilePointer(_BasePointer):
-    """A pointer analogous to a file offset within the Mach-O
-    """
+    """A pointer analogous to a file offset within the Mach-O."""
 
     def __str__(self) -> str:
         return f"Phys[{super().__str__()}]"
@@ -41,12 +40,11 @@ class StaticFilePointer(_BasePointer):
 
 
 class VirtualMemoryPointer(_BasePointer):
-    """A pointer representing a virtual memory location within the Mach-O
-    """
+    """A pointer representing a virtual memory location within the Mach-O."""
 
 
 def swap32(i: int) -> int:
-    """Reverse the bytes of a little-endian integer representation ie (3) -> 50331648"""
+    """Reverse the bytes of a little-endian integer representation ie (3) -> 50331648."""
     return struct.unpack("<I", struct.pack(">I", i))[0]
 
 
@@ -63,6 +61,9 @@ class MachArch(IntEnum):
     MH_CPU_TYPE_ARM = 12
     MH_CPU_TYPE_ARM64 = MH_CPU_TYPE_ARM | MH_CPU_ARCH_ABI64
 
+    MH_CPU_SUBTYPE_PTRAUTH_ABI = 0x80000000
+    MH_CPU_SUBTYPE_ARM64E = 0x02
+
     DYLD_SHARED_CACHE_MAGIC = 0x646C7964  # b'dyld'
 
 
@@ -672,3 +673,143 @@ class CFString32(Structure):
 
 class CFString64(Structure):
     _fields_ = [("base", c_uint64), ("flags", c_uint64), ("literal", c_uint64), ("length", c_uint64)]
+
+
+class MachoDyldChainedFixupsHeaderRaw(Structure):
+    _fields_ = [
+        # 0
+        ("fixups_version", c_uint32),
+        # Offset of dyld_chained_starts_in_image in chain_data
+        ("starts_offset", c_uint32),
+        # Offset of imports table in chain_data
+        ("imports_offset", c_uint32),
+        # Offset of symbol strings in chain_data
+        ("symbols_offset", c_uint32),
+        # Number of imported symbol names
+        ("imports_count", c_uint32),
+        # DYLD_CHAINED_IMPORT*
+        ("imports_format", c_uint32),
+        # 0 => uncompressed, 1 => zlib compressed
+        ("symbols_format", c_uint32),
+    ]
+
+
+class MachoDyldChainedImportFormat(IntEnum):
+    DYLD_CHAINED_IMPORT = 1
+    DYLD_CHAINED_IMPORT_ADDEND = 2
+    DYLD_CHAINED_IMPORT_ADDEND64 = 3
+
+
+class MachoDyldChainedImportRaw(Structure):
+    _fields_ = [
+        ("lib_ordinal", c_uint32, 8),
+        ("weak_import", c_uint32, 1),
+        ("name_offset", c_uint32, 23),
+    ]
+
+
+class MachoDyldChainedImportAddend64Raw(Structure):
+    _fields_ = [
+        ("lib_ordinal", c_uint64, 16),
+        ("weak_import", c_uint64, 1),
+        ("reserved", c_uint64, 15),
+        ("name_offset", c_uint64, 32),
+        ("addend", c_uint64),
+    ]
+
+
+class MachoDyldChainedStartsInImageRaw(Structure):
+    _fields_ = [
+        ("seg_count", c_uint32),
+        # Each entry is offset into this struct for that segment
+        # followed by pool of dyld_chain_starts_in_segment data
+        # XXX(PT): Although this is declared as uint32_t[1] array in the dyld source, it's actually
+        # an array of `seg_count` entries.
+        # To avoid hacks that would let ctypes parse this correctly, we manually read the offsets
+        # ("seg_info_offset", c_uint32 * 1),
+    ]
+
+
+class MachoDyldChainedStartsInSegmentRaw(Structure):
+    # XXX(PT): Force alignment to uint16_t, as by default this structure is reported as being 2 bytes too big
+    # The correct size is important when we're parsing data that is placed directly after this structure
+    # https://docs.python.org/3/library/ctypes.html#structure-union-alignment-and-byte-order
+    _pack_ = 2
+
+    _fields_ = [
+        # Size of this (amount kernel needs to copy)
+        ("size", c_uint32),
+        # 0x1000 or 0x4000
+        ("page_size", c_uint16),
+        # MachoDyldChainedPtrFormat
+        ("pointer_format", c_uint16),
+        # Offset in memory to start of segment
+        ("segment_offset", c_uint64),
+        # For 32-bit OS, any value beyond this is not a pointer
+        ("max_valid_pointer", c_uint32),
+        # How many pages are in array
+        ("page_count", c_uint16),
+        # Each entry is offset in each page of first element in chain
+        # or DYLD_CHAINED_PTR_START_NONE if no fixups on page
+        # XXX(PT): Variable-length array, see comment on MachoDyldChainedStartsInImageRaw
+        # ("page_start", c_uint16),
+        # Some 32-bit formats may require multiple starts per page.
+        # For those, if high bit is set in page_starts[], then it
+        # is index into chain_starts[] which is a list of starts
+        # the last of which has the high bit set
+        # XXX(PT): Variable-length array, see comment on MachoDyldChainedStartsInImageRaw
+        # ("chain_starts", c_uint16),
+    ]
+
+
+class MachoDyldChainedPointerStartType(IntEnum):
+    # Used in page_start[] to denote a page with no fixups
+    DYLD_CHAINED_PTR_START_NONE = 0xFFFF
+    # Used in page_start[] to denote a page which has multiple starts
+    DYLD_CHAINED_PTR_START_MULTI = 0x8000
+    # Used in chain_starts[] to denote last start in list for page
+    DYLD_CHAINED_PTR_START_LAST = 0x8000
+
+
+class MachoDyldChainedPtrFormat(IntEnum):
+    """Gives the format for packed pointers within fixup chains
+    PT: Only model a subset for now so it's clear when we encounter binaries breaking our assumptions
+    Ref: https://opensource.apple.com/source/dyld/dyld-851.27/include/mach-o/fixup-chains.h.auto.html
+    """
+
+    # Packed target is a virtual memory address
+    DYLD_CHAINED_PTR_64 = 2
+    # Packed target is an offset from the runtime virtual base
+    DYLD_CHAINED_PTR_64_OFFSET = 6
+
+
+class MachoDyldChainedPtr64RebaseRaw(Structure):
+    # Used with DYLD_CHAINED_PTR_64/DYLD_CHAINED_PTR_64_OFFSET
+    _fields_ = [
+        # Interpreted according to the active pointer format
+        # Note that 36 bits imposes a maximum image size of 64GB
+        ("target", c_uint64, 36),
+        # Top 8 bits set to this (DYLD_CHAINED_PTR_64 => after slide added,
+        # DYLD_CHAINED_PTR_64_OFFSET => before slide added)
+        ("high8", c_uint64, 8),
+        # All zeros
+        ("reserved", c_uint64, 7),
+        # 4-byte stride
+        ("next", c_uint64, 12),
+        # == 0
+        ("bind", c_uint64, 1),
+    ]
+
+
+class MachoDyldChainedPtr64BindRaw(Structure):
+    _fields_ = [
+        ("ordinal", c_uint64, 24),
+        # 0 thru 255
+        ("addend", c_uint64, 8),
+        # All zeroes
+        ("reserved", c_uint64, 19),
+        # 4-byte stride
+        ("next", c_uint64, 12),
+        # == 1
+        ("bind", c_uint64, 1),
+    ]
